# ========================================================================================================
# Assetto Corsa Tyre Buddy
# - written by Becky York - May 2022
# Designed to help optimise tyre pressures, brake bias & camber
# ========================================================================================================
# v0.0.1 alpha
# Inside/middle/outside tyre temps
# Ensure IMO temps are within +/- 10C and that M is ~50% of the difference between I and O
# v0.1.0 alpha
# Complete re-write
# Added CSP compatible config
# IMO & Core Colours are now calculated using LUTs
# refactor tyres GUI: scale of 0 to 1 & multiply that by window widthXheight
# Bars are bolder so they can be seen without squinting
# Added core temp guage (75% of core temp + 25% of IMO temp)
# Added PSI indicator (black inside the core temp guage)
#
# v0.9.0 beta - 2023-Aug-08
# Added Optimal Tyre Temperatures for each CAR/COMPOUND combination so GREEN really means GOOD!
# Implemented Tyre Stability which monitors tyre temperature over 2kms and puts up a warning if it exceeds 5°C difference
# Another substantial re-write
# Implementation of MVC seperation of duties
# Added PSI Recommendations per tyre
# Added Slips difference indicator between the right & left side of the car
#  -- GREEN means left and right tyres are doing the same amount of steering
#  -- RED means left and right tyres are trying to turn the car different amounts
# Added FIRE/HOT/COLD Icons
# Added YELLOW Thermal Stability - this affects all tyres and means you have to drive until the counter is zero
# Added ORANGE Thermal Stability - this turns on when individual tyres get thermal shock due to sliding/braking
# Added Text Readouts - NUMBERS!
# - Page -1 : off
# - Page  0 : Auto  - this will pick the best page 1-3 based on what's going on with the tyres
# - Page  1 : TEMPS - Optimal Range, Current Temperature, Temperature Variation over the last 2KM
# - Page  2 : PSI   - Current Temperature, Temperature Variation : Recommended PSI
# - Page  3 : IMO   - IM MO IMO Temperature Spread
# Added Optimal Spread Settinngs; 2.5,5.0,7.5,10.0,12.5°C this should account for most car requirements
#
# v1.0.0 - 2023-Aug-22
# Camber: Aggressivo / Equilibrato / Freddo / Super Freddo
# Changed IMO "Over-Camber" Bar Colour Gradient from percentage to fixed value
# CSP Setting: 5/10/15/20km tyre warmup   (default 15)
# CSP Setting: 2/ 5/ 7/10km spin recovery (default 2)
# Removed potential namespace collisions with other python apps
#
# v1.1.0 - 2023-Aug-30
# Make the App size changable 720p, 1080p, 1440p, 4k

# v1.2.0 - 2023-Sept-22
# Added Tyre Wear

# v1.3.0 - 2023-Nov-17
# Added Fuel Meter
# Litres Per Lap & Litres Per Minute
# Estimated Laps & Minutes until fuel runs dry
# Litres to put in at next pit stop
# Litres required to finish session

# TODO
# Rear Lights - instead of toe and diff in slip angle.
# --- make it difference in wheel speed left/right with center biased green bits to indicate which tyre is slipping faster
# DEMO MODE - tell people how to use the damn thing
# Per car settings override
# apps icon (on/off)
# BRAKE BIAS
# when moving from not braking to braking - record tyre temps
# during braking - record tyre temps
# when finished braking - record tyre temps
# compare front/rear temperature differences between start and end of braking

import ac
import acsys
import sys
import os
import platform

try:
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), "TyreBuddy_dll_x64"))
    os.environ['PATH'] = os.environ['PATH'] + ";." 
except:
    exc_type, exc_value, exc_traceback = sys.exc_info()
    ac.log("TyreBuddy: Failed to import _cTypes")

from TyreBuddy_lib.sim_info import info
from TyreBuddy_lib.config import Config
from TyreBuddy_lib.CAC import cCAC
from TyreBuddy_lib.buddy_app import cBuddy_App
from TyreBuddy_lib.buddy_drive import cBuddy_Drive
from TyreBuddy_lib.buddy_tyres import cBuddy_Tyres
from TyreBuddy_lib.buddy_fuel  import cBuddy_Fuel
from TyreBuddy_lib.beancounter import cBeanCounter

class cTyreBuddy:
    def __init__(self):
        self.config = Config()
        self.loaded = 0
        self.shutdown = 0
        self.deltaT = 0
        self.ACConstants = cCAC()
        self.UpdateBean = cBeanCounter(5)
        self.BuddyDrive = cBuddy_Drive(self.config)
        self.BuddyDrive.RegisterEvent("PitLane_Enter"  ,self.BuddyDrive_PitLane_Enter)
        self.BuddyDrive.RegisterEvent("PitLane_Exit"   ,self.BuddyDrive_PitLane_Exit)
        self.BuddyDrive.RegisterEvent("Box_Enter"      ,self.BuddyDrive_Box_Enter)
        self.BuddyDrive.RegisterEvent("Box_Exit"       ,self.BuddyDrive_Box_Exit)
        self.BuddyDrive.RegisterEvent("New_Lap"        ,self.BuddyDrive_New_Lap)
        self.BuddyDrive.RegisterEvent("Driving_Started",self.BuddyDrive_Driving_Started)
        self.BuddyDrive.RegisterEvent("Driving_Stopped",self.BuddyDrive_Driving_Stopped)
        self.BuddyDrive.RegisterEvent("Driving"        ,self.BuddyDrive_Driving)
        self.BuddyDrive.RegisterEvent("New_Tyres"      ,self.BuddyDrive_New_Tyres)
        self.BuddyTyres = cBuddy_Tyres(self.config)
        self.BuddyFuel  = cBuddy_Fuel()
        
        self.TextBean = cBeanCounter(15)

        self.BuddyApp = cBuddy_App(self.config)

    def startup(self):
        self.BuddyApp.createWindow()
        self.BuddyApp.setup(self)
        
        lOptimalSpread_front, lOptimalSpread_rear = self.config.IMOSpreads.Spread
        self.BuddyTyres.updateOptimalSpread(lOptimalSpread_front,lOptimalSpread_rear)

        self.loaded = 1

    def Shutdown(self):
        self.shutdown = 1

    def BuddyDrive_PitLane_Enter(self,DictArgs):
        self.BuddyFuel.InPits()

    def BuddyDrive_PitLane_Exit(self,DictArgs):
        #even if you didn't change tyres, they have cooled
        self.BuddyDrive.ResetDistanceRequired(0,1)
        isTouristenfahrten = self.config.isTouristenfahrten
        if isTouristenfahrten:
            self.BuddyFuel.LapStart()

    def BuddyDrive_Box_Enter(self,DictArgs):
        return

    #when the car exits the box, check for fresh tyres
    def BuddyDrive_Box_Exit(self,DictArgs):
        self.BuddyDrive.CheckTyres()

    def BuddyDrive_New_Lap(self,DictArgs):
        isTouristenfahrten = self.config.isTouristenfahrten
        self.BuddyFuel.LapEnd(isTouristenfahrten)

    #if the tyres have changed on the car
    #then load the new config for the car/compound
    #reset the distance required to 10km
    def BuddyDrive_New_Tyres(self,DictArgs):
        self.config.CalculateTyres(DictArgs["CarName"],DictArgs["Compound"])
        self.BuddyDrive.ResetDistanceRequired(1,0)
        self.BuddyTyres.NewTyres()

    def BuddyDrive_Driving_Started(self,DictArgs):
        return

    #when the car stops, reset the distance required to 2km
    def BuddyDrive_Driving_Stopped(self,DictArgs):
        self.BuddyDrive.ResetDistanceRequired(0,1)
        # Ensure that tyres get an update even though we have stopped
        self.BuddyDrive_Driving(DictArgs)

    #This event gets triggered every ~100 meters
    def BuddyDrive_Driving(self,DictArgs):
        lDrivingRequired = 1 if self.BuddyDrive.Distance_KMsToGo > 0 else 0
        self.BuddyTyres.UpdateAVG_2km(lDrivingRequired)

    def update(self, deltaT):
        if not self.UpdateBean:
            return
        self.deltaT = deltaT

        self.BuddyDrive.update()
        self.BuddyTyres.update()

    def draw(self):
        #self.BuddyApp.DrawElementPositions()
        self.BuddyApp.DrawTyres()
        self.BuddyApp.DrawSlips()

        if self.TextBean:
            self.BuddyApp.write_debug_text()

#create new TyreBuddy class instance
TyreBuddy = cTyreBuddy()

def acMain(ac_version):
    global TyreBuddy
    TyreBuddy.startup()
    
    ac.addRenderCallback(TyreBuddy.BuddyApp.mainWindow , acDoDraw)
    return TyreBuddy.config.app_name

def acDoDraw(deltaT):
    #add update limits to reduce cpu cycles
    global TyreBuddy
    TyreBuddy.draw()

def acUpdate(deltaT):
    #add update limits to reduce cpu cycles
    global TyreBuddy
    TyreBuddy.update(deltaT)

def acShutdown(*args):
    global TyreBuddy
    TyreBuddy.Shutdown()

